/*
 *	YZ - Print the tree structure of a directory
 *	Usage:	yz [-dD] [top]
 *
 *	-d	print only directorys
 *	-D	print datastructure
 */
#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<pwd.h>
#include	<sys/dir.h>
#define	DEFDIRSIZ	14
#define DEFCOLWID       14
#ifndef MAXNAMLEN
# define MAXNAMLEN      DEFCOLWID 
#endif
/*
 * TWIDDLE is a fudge factor.  It should be declared so that
 * sizeof(struct entry) is on a nice boundry.
 */
#define	TWIDDLE	4
/*	addr is defined as a function to compensate for a compiler bugg !!
 *	define	addr(b,o) ((struct entry *)
 *			((int)(b) + (o)*(sizeof(struct entry)+Length-TWIDDLE)))
 */
#define	SIZEOFentry	(sizeof(struct entry) + Length - TWIDDLE)
#define	DEPTH	17	/* maximum depth that yz will go */
#define	MAX	100	/* initial # of elements for list of files */
#define	PWD	"/bin/pwd"	/* program to get name of current dir */
#define	FFIRST	2	/* sort files first */
#define	FAIL	-1	/* failure return status of sys calls */
#define	GREATER	1	/* return value of strcmp if arg1 > arg2 */
#define	LESSTHAN -1	/* return value of strcmp if arg1 < arg2 */
#define	SAME	0	/* return value of strcmp if arg1 == arg2 */
int	Index = 0;		/* current element of list[]	*/
int	Length = DEFDIRSIZ;	/* max length of a dir. name    */
int	All = 1;		/* list non directory entries	*/
int	Debug = 0;		/* no debug			*/
int	File_dir = FFIRST;	/* flag for how to sort		*/
int	Maxes[DEPTH];		/* array keeps track of max length in columns */
int	Level = 0;		/* counter for how deep we are	*/
int	Device;			/* device that we are starting tree on */
struct	stat Status;
char	Array[DEPTH][256];	/* memory is cheap 	*/
struct entry {
	int next;			/* index to next element in list */
					/* could be a ptr, but realloc() */
					/* might screw us then */
	off_t	e_size;			/* size in blocks */
	unsigned short	e_mode;		/* file mode		*/
	
	short unsigned dir : 1;		/* entry is a directory */
	short unsigned last : 1;	/* last entry in the dir. */
	short unsigned dev : 1;		/* set if same device as top */
	short unsigned end : 13;	/* index of last subdir entry*/
	char	e_name[TWIDDLE];	/* name from directory entry */
					/* it will actually be larger */
} *List, *SaveList;
unsigned Size;				/* how big of space we've malloced */
char	*Spaces;	/* used for output */
char	Buf1[BUFSIZ];	/* buffers for stdio stuff.  We don't want	*/
char	Buf2[BUFSIZ];	/* anyone calling malloc, because then		*/
char	Buf3[BUFSIZ];	/* realloc() will have to move the whole list	*/

main(argc, argv)

char	**argv;
int	argc;
{
    register int    i, j;
    char    top[128];		/* array for treetop name */
    char    home[128];		/* starting dir for multiple trees */
    char   *ptr;
    char   *malloc ();
    char   *rindex ();
    FILE * istr, *popen ();
    struct entry   *l;
    struct entry   *addr ();

    setbuf (stdout, Buf1);
    for (j = 1; j < argc; j++) {
	if (argv[j][0] == '-') {
	    for (i = 1; i < strlen (argv[j]); i++) {
		switch (argv[j][i]) {
		    case 'd': 
			All = 0;
			break;
		    case 'D': 
			Debug = 1;
			break;
		    default: 
			fprintf (stderr, "Usage: yz [-dD] [directory ... ]\n");
			exit (FAIL);
		}
	    }
	}
	else
	    break;
    }
    /* Establish where we are (our home base...) */
    if ((istr = popen (PWD, "r")) == NULL) {
	fprintf (stderr, "yz: %s failed\n", PWD);
	exit (1);
    }
    else {
	setbuf (istr, Buf2);
	if (fgets (home, 128, istr) == NULL) {
	    fprintf (stderr, "yz: EOF from %s\n", PWD);
	    exit (1);
	}
    }
    pclose (istr);
    if (home[strlen (home) - 1] == '\n')
	home[strlen (home) - 1] = '\0';
    Spaces = malloc (Length + 2);
    for (i = 0; i < Length + 1; i++)
	Spaces[i] = ' ';
    Spaces[i] = '\0';
    Size = SIZEOFentry * MAX;	/* Get initial Storage space */
    SaveList = (struct entry   *) malloc (Size);
    if (j == argc)    /* adjust for no specified directory */
	argv[--j] = ".\0";
    /* walk down the rest of the args, treeing them one at at time */
    for (; j < argc; j++) {
	if (chdir (home) == -1)	{
	    fprintf (stderr, "Can't chdir back to %s\n", home);
	    exit (1);
	}
	sprintf (top, "%s", argv[j]);
	if (chdir (top) == FAIL) {
	    fprintf (stderr, "Can't chdir to %s\n", top);
	    continue;
	}
	else
	    if ((istr = popen (PWD, "r")) == NULL) {
		fprintf (stderr, "yz: %s failed\n", PWD);
		continue;
	    }
	else {
	    setbuf (istr, Buf2);
	    if (fgets (top, 128, istr) == NULL) {
		fprintf (stderr, "yz: EOF from %s\n", PWD);
		pclose (istr);
		continue;
	    }
	    ptr = top;
	    while (*ptr && (*ptr != '\n'))
		ptr++;
	    *ptr = '\0';
	    pclose (istr);
	}
	List = SaveList;
	Index = 0;
	ptr = rindex (top, '/');
	if (!ptr || *++ptr == '\0')
	    sprintf (addr (List, Index) -> e_name, "%.*s", Length, top);
	else
	    sprintf (addr (List, Index) -> e_name, "%.*s", Length, ptr);

	if (stat (top, &Status) == FAIL) {
	    fprintf (stderr, "Can't stat %s\n", top);
	    continue;
	}
	Device = Status.st_dev;
	addr (List, 0) -> dir = 1;
	addr (List, 0) -> last = 1;
	addr (List, 0) -> next = 1;
	addr (List, 0) -> e_mode = Status.st_mode;
	addr (List, 0) -> e_size = Status.st_size;
	Index = 1;
	for (i = 1; i < DEPTH; i++)
	    Maxes[i] = 0;
	Maxes[0] = stln (addr (List, 0) -> e_name);
	Level = 1;
	/* search the tree */
	addr (List, 0) -> end = t_search (top, addr (List, 0));
	if (Index == 1)		/* empty tree */
	    addr (List, 0) -> next = 0;
	if (Debug == 1) {	/* print the list	*/
 	    fflush(stdout);
	    for (i=0;strcmp(addr(List,i)->e_name, "") != 0;i++) {
		l = addr(List,i);
		printf("%3d %3d %14.14s" , i, l->next, l->e_name);
		printf("%7lu ", l->e_size);
		printf("%2u %2u %2u %2u ", l->dir,l->last, l->dev, l->end);
		printf("\n");
	    }
	    fflush(stdout);
	}
	for (i = addr(List, 0) -> next;(int) addr (List, i) -> last == 0; i++);
	size_patch (addr (List, 0) -> next, i);
	for (i = addr(List, 0) -> next;(int) addr (List, i) -> last == 0; i++);
	pr_list (addr (List, 0) -> next, i);
    }
    exit (0);
}

size_patch(first,last)

int	first,last;
{
    struct entry   *l;
    int     i, j;
    off_t size_rec ();

    /* this is the main loop to patch the size member in structure * entry for
       a directory. */
    for (j = l -> next; (int) addr (List, j) -> last == 0; j++);
    l -> e_size = size_rec (l -> next, j);
}

off_t size_rec(first,last)

int	first,last;
{
    int     i, j;
    struct entry   *l;
    off_t summ;

    summ = 0;
    for (i = first; i <= last; i++) {
	l = addr (List, i);
	if (l -> dir == 1) {
	    if (l -> next != 0) {
		for (j = l -> next; (int) addr (List, j) -> last == 0; j++);
		l -> e_size = size_rec (l -> next, j);
		summ += l -> e_size;
	    }
	    else l -> e_size = 0;
	}
	else summ += l -> e_size;
    }
    return (summ);
}

pr_list(first,last)

int	first,last;
{
    struct entry   *l;
    int     i, j;

    /* this is the main loop to print the tree structure with *
       dimensional-flowshart style. */
    strcpy (Array[0], "|");
    printf ("|\n|\\\n");
    for (j = l -> next; (int) addr (List, j) -> last == 0; j++);
    pr_rec (l -> next, j, 0);
}

pr_rec(first,last,level)

int	first,last,level;
{
    int     i, j, k;
    int	    moredir;
    struct entry   *l;

    for (i = first; i <= last; i++) {
	l = addr (List, i);
	if (l -> next != 0 || l -> dir == 1) {
	    printf ("%s | %-s   ", Array[level], l -> e_name);
	    printf ("%u", l -> e_size);
	    printf ("\n");
	}
	else {
	    if (All) {
		printf ("%s | %-14s ", Array[level], l -> e_name);
		printf ("%7lu", l -> e_size);
		printf ("\n");
	    }
	}
	if (l -> next != 0) {
	    moredir = 0;
	    for (j = l -> next; (int) addr (List, j) -> last == 0; j++);
	    for (k = l -> next; k <= j; k++) {
		if (addr(List, k) -> dir == 1) {
		    moredir = 1;
		}
	    }
	    if ( All || moredir) {
		strcpy (Array[level + 1], Array[level]);
		if (i == last)
		    strcat (Array[level + 1], "   ");
		else
		    strcat (Array[level + 1], " | ");
		for (k = 0; k < strlen (l -> e_name); k++)
		    strcat (Array[level + 1], " ");
		printf ("%s\\\n", Array[level + 1]);
	    }
	    pr_rec (l -> next, j, level + 1);
	}
    }
}

struct entry *

addr(b,o)

struct entry *b;
int o;
{
    int     x;

    x = sizeof (struct entry) + Length - TWIDDLE;
    return ((struct entry  *) ((int) (b) + (o * x)));
}

t_search(dir, addrs)

char *dir;
struct	entry *addrs;
{
    int     bsort;		/* index to begin sort */
    int     stmp;		/* save temporary index value */
    struct entry   *sstep;	/* saved step in list */
    int     nitems;		/* # of items in this directory */
    FILE * dirp;		/* pointer to directory */
    char    sub[MAXNAMLEN + 1];	/* used for subdirectory names */
    int     i;
    struct direct   dirent;
    struct direct  *dp = &dirent;
    int     n_subs = 0;
    int     tmp = 0;
    extern  qsort ();
    int     compar ();		/* comparison routine for qsort */
    char   *realloc ();

    dirp = fopen (".", "r");
    if (dirp == NULL) {
	fprintf (stderr, "Cannot open %s\n", dir);
	return (0);
    }
    setbuf (dirp, Buf3);
    bsort = Index;
    sstep = addr (List, bsort);	/* initialize sstep for for loop later on */
    nitems = Index;
    /* get the entries of the directory that we are interested in */
    while (fread ((char *) (dp), sizeof (struct direct), 1, dirp) == 1) {
	if (!dp -> d_ino
	    || (strcmp (dp -> d_name, ".") == SAME)
	    || (strcmp (dp -> d_name, "..") == SAME))
	    continue;
	sprintf (sub, "%s", dp -> d_name);
	if ((tmp = stat (sub, &Status)) == FAIL) {
	    fprintf (stderr, "%s:stat can't find\n", sub);
	    continue;
	}
	else
	    if ((Status.st_mode & S_IFMT) == S_IFDIR)
                addr (List, Index) -> dir = 1;
	    else addr (List, Index) -> dir = 0;
	sprintf (addr (List, Index) -> e_name, "%.*s", Length, dp -> d_name);
	addr (List, Index) -> last = 0;
	addr (List, Index) -> end = 0;
	addr (List, Index) -> dev = (Device == Status.st_dev);
	addr (List, Index) -> e_mode = Status.st_mode;
	addr (List, Index) -> e_size = Status.st_size;
	if (stln (addr (List, Index) -> e_name) > Maxes[Level])
	    Maxes[Level] = stln (addr (List, Index) -> e_name);
	++Index;
	if (Index * SIZEOFentry >= Size) {
	    Size += 20 * SIZEOFentry;
	    if (!(List = (struct entry  *) realloc ((char *) List, Size))) {
		fprintf (stderr, "Out of space\n");
		break;
	    }
	}
    }
    fclose (dirp);
    nitems = Index - nitems;	/* nitems now contains the # of */
    /* items in this dir, rather than */
    /* # total items before this dir */
    qsort (addr (List, bsort), nitems, SIZEOFentry, compar);
    addr (List, Index - 1) -> last = 1;/* mark last item for this dir */
    n_subs = nitems;
    stmp = Index;
    /* now walk through, and recurse on directory entries */
    /* sstep was initialized above */
    for (i = 0; i < nitems; sstep = addr (List, stmp - nitems + (++i))) {
	if (sstep -> dir) {
	    sstep -> next = Index;
	    sprintf (sub, "%.*s", Length, sstep -> e_name);
	    tmp = n_subs;
	    Level++;
	    if (chdir (sub) == FAIL)
		fprintf (stderr, "Can't chdir to %s/%s\n", dir, sub);
	    else {
		n_subs += t_search (sub, sstep);
		if (chdir ("..") == FAIL) {
		    fprintf (stderr, "No '..' in %s/%s!\n", dir, sub);
		    exit (1);
		}
	    }
	    --Level;
	    if (n_subs - tmp <= 0)
		sstep -> next = 0;
	    else --n_subs;
	}
	else sstep -> next = 0;
    }
    addrs -> end = (unsigned) n_subs;
    return (n_subs);
}
/*
 *	comparison routine for qsort
 */
compar(a, b)

struct entry *a, *b;
{
    if (!File_dir)		/* straight alphabetical */
	return (strncmp (a -> e_name, b -> e_name, Length));
    /* sort alphabetically if both dirs or both not dirs */
    if ((a -> dir && b -> dir) || (!a -> dir && !b -> dir))
	return (strncmp (a -> e_name, b -> e_name, Length));
    if (a -> dir)
	 return (GREATER);
    else return (LESSTHAN);
}

/* stln - sortof like strlen, returns length up to Length-1 */

stln(st)

register char *st;
{
    register int    t;

    for (t = 0; t < Length - 1; ++t)
	if (!st[t])
	    return (++t);
    return (++t);
}
/*
 * Return a pointer into str at which the character
 * c last appears; NULL if not found.
 */
char *
rindex(str, c)

register char *str, c;
{
    register char  *ptr;

    ptr = NULL;
    do {
	if (*str == c)
	    ptr = str;
    } while (*str++);
    return (ptr);
}
